@@ -3,7 +3,7 @@ package Promise::ES6;
 use strict;
 use warnings;
 
-our $VERSION = '0.18_01';
+our $VERSION = '0.19';
 
 =encoding utf-8
 
@@ -72,24 +72,17 @@ L<indicates intent|https://www.ecma-international.org/ecma-262/6.0/#sec-promise-
 
 =head1 COMPATIBILITY
 
-Right now this doesn’t try for interoperability with other promise
-classes. If that’s something you want, make a feature request.
+This module considers any object that has a C<then()> method to be a promise.
+Note that, in the case of L<Future>, this will yield a “false-positive”, as
+Future is not compatible with promises.
 
-See L<Promise::ES6::Future> if you need to interact with L<Future>.
+(See L<Promise::ES6::Future> for more tools to interact with L<Future>.)
 
 =head1 UNHANDLED REJECTIONS
 
-As of version 0.05, unhandled rejections prompt a warning I<only> if one
-of the following is true:
-
-=over
-
-=item 1) The unhandled rejection happens outside of the constructor.
-
-=item 2) The unhandled rejection happens via an uncaught exception
-(even within the constructor).
-
-=back
+This module’s handling of unhandled rejections has changed over time.
+The current behavior is: if any rejected promise is DESTROYed without first
+having received a failure callback, a warning is thrown.
 
 =head1 SYNCHRONOUS OPERATION
 
@@ -118,9 +111,9 @@ A key advantage of this design is that Promise::ES6 instances can abstract
 over whether a given function works synchronously or asynchronously.
 
 If you want a Promises/A+-compliant implementation, look at
-L<Promise::ES6::IOAsync>, L<Promise::ES6::Mojo>,
-L<Promise::ES6::AnyEvent>, or one of the alternatives
-that that module’s documentation suggests.
+L<Promise::ES6::IOAsync>, L<Promise::ES6::Mojo>, or
+L<Promise::ES6::AnyEvent> in this distribution. CPAN provides other
+alternatives.
 
 =head1 CANCELLATION
 
@@ -201,23 +194,21 @@ You may also (counterintuitively, IMO) find that this:
 
 =back
 
-=head1 TODO
-
-Currently rejections will defer until promises are resolved. This makes
-no real sense and ought to change. Do not build anything that depends on
-this behavior.
-
 =head1 SEE ALSO
 
 If you’re not sure of what promises are, there are several good
 introductions to the topic. You might start with
 L<this one|https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Using_promises>.
 
-Promise::ES6 serves much the same role as L<Future> but exposes
-a standard, minimal, cross-language API rather than a proprietary (large) one.
+L<Promise::XS> is a lot like this library but implemented mostly in XS for
+speed. It derives from L<AnyEvent::XSPromises>.
+
+L<Promises> is another pure-Perl Promise implementation.
+
+L<Future> fills a role similar to that of promises.
 
 CPAN contains a number of other modules that implement promises. I think
-mine is the nicest :), but YMMV. Enjoy!
+mine are the nicest :), but YMMV. Enjoy!
 
 =head1 LICENSE & COPYRIGHT
 
@@ -233,8 +224,6 @@ our $DETECT_MEMORY_LEAKS;
 
 sub catch { $_[0]->then( undef, $_[1] ) }
 
-sub finally { $_[0]->then( $_[1], $_[1] ) }
-
 sub resolve {
     my ( $class, $value ) = @_;
 
@@ -273,12 +262,10 @@ sub all {
                 };
 
                 for my $promise (@promises) {
-                    last if $settled;
-
                     my $p = $p++;
 
                     $promise->then(
-                        sub {
+                        $settled ? undef : sub {
                             return if $settled;
 
                             $values[$p] = $_[0];
@@ -337,8 +324,6 @@ sub race {
     };
 
     for my $promise (@promises) {
-        last if $is_done;
-
         $promise->then( $on_resolve_cr, $on_reject_cr );
     }
 
